/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.hyracks.control.nc.io.cloud;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;

import org.apache.commons.io.FileUtils;
import org.apache.hyracks.api.exceptions.HyracksDataException;
import org.apache.hyracks.api.io.FileReference;
import org.apache.hyracks.api.io.IFileHandle;
import org.apache.hyracks.api.io.IIOManager;
import org.apache.hyracks.control.nc.io.cloud.clients.ICloudClient;
import org.apache.hyracks.util.file.FileUtil;

// TODO replace with a proper caching mechanism
public class LocalCacheUtil {
    private LocalCacheUtil() {

    }

    public static void writeToFile(FileReference fileRef, byte[] bytes) throws HyracksDataException {
        try {
            File file = fileRef.getFile();
            FileUtils.createParentDirectories(file);
            FileUtil.writeAndForce(file.toPath(), bytes);
        } catch (IOException e) {
            throw HyracksDataException.create(e);
        }
    }

    // TODO: replace with proper caching policy
    public static void download(ICloudClient cloudClient, CloudIOManager ioManager, CloudFileHandle fileHandle,
            IIOManager.FileReadWriteMode rwMode, IIOManager.FileSyncMode syncMode, ByteBuffer writeBuffer)
            throws HyracksDataException {
        FileReference fileRef = fileHandle.getFileReference();
        // write the file locally (the call to open is synchronized hence only one thread can perform this call)
        try (InputStream inStream = cloudClient.getObjectStream(ioManager.getBucket(), fileRef.getRelativePath())) {
            File file = fileRef.getFile();
            FileUtils.createParentDirectories(fileRef.getFile());
            if (!file.createNewFile()) {
                throw new IllegalStateException("Couldn't create local file");
            }
            fileHandle.open(rwMode, syncMode);
            LocalCacheUtil.writeToFile(ioManager, fileHandle, inStream, writeBuffer);
        } catch (IOException e) {
            throw HyracksDataException.create(e);
        }
    }

    private static void writeToFile(CloudIOManager ioManager, IFileHandle fileHandle, InputStream inStream,
            ByteBuffer writeBuffer) throws HyracksDataException {
        writeBuffer.clear();
        try {
            int position = 0;
            long offset = 0;
            int read;
            while ((read = inStream.read(writeBuffer.array(), position, writeBuffer.remaining())) >= 0) {
                position += read;
                writeBuffer.position(position);
                if (writeBuffer.remaining() == 0) {
                    offset += writeBufferToFile(ioManager, fileHandle, writeBuffer, offset);
                    position = 0;
                }
            }

            if (writeBuffer.position() > 0) {
                writeBufferToFile(ioManager, fileHandle, writeBuffer, offset);
                ioManager.syncLocally(fileHandle);
            }
        } catch (IOException e) {
            throw HyracksDataException.create(e);
        }

    }

    private static long writeBufferToFile(CloudIOManager ioManager, IFileHandle fileHandle, ByteBuffer writeBuffer,
            long offset) throws HyracksDataException {
        writeBuffer.flip();
        long written = ioManager.writeLocally(fileHandle, offset, writeBuffer);
        writeBuffer.clear();
        return written;
    }
}
